{% extends "base.html" %}
{% block title %}Initializing{% end %}
{% block content %}
<div id="initializing">
  <p>$ load "imbi.js", 8, 1<p>
  <p class="italic">
    Initializing Imbi v{{ handler.application.settings['version'] }} <span class="blink">&hellip;</span>
  </p>
</div>
{% end %}
